@@ -2,7 +2,7 @@
    pstoedit.cpp : This file is part of pstoedit
    main control procedure
 
-   Copyright (C) 1993 - 2023 Wolfgang Glunz, wglunz35_AT_pstoedit.net
+   Copyright (C) 1993 - 2024 Wolfgang Glunz, wglunz35_AT_pstoedit.net
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -64,6 +64,10 @@ PsToEditOptions& PsToEditOptions::theOptions() // singleton
 
 #include "psfront.h"
 
+#if !defined(_MSC_VER) && !defined(__CYGWIN__)
+// for iterating over shared libraries
+#include <link.h>
+#endif
 
 extern const char *defaultPIoptions(ostream & errstream, int verbose);	// in callgs.cpp
 
@@ -159,6 +163,7 @@ private:	// yes these are private, library users should use the public interface
 	void show_path() override {	}
 	void show_rectangle(const float /* llx */ , const float /* lly */ , const float /* urx */ ,
 						const float /* ury */ ) override {	}
+	NOCOPYANDASSIGN(drvNOBACKEND)
 };
 // *INDENT-ON*
 
@@ -270,6 +275,17 @@ ProgramOptions* getProgramOptionsForDriver(const char* driverName) {
 
 
 #ifndef UPPVERSION
+#if !defined(_MSC_VER) && !defined(__CYGWIN__)
+static int dl_iterate_callback(struct dl_phdr_info *info, size_t size, void *data){
+	if (size && info && strstr(info->dlpi_name, "libpstoedit.so.")) {
+//		cout << "found libpstoedit.so in " << info->dlpi_name << endl;
+		*((RSString*)data) = info->dlpi_name;
+		return 1;
+	}
+	return 0;
+}
+#endif
+
 extern "C" DLLEXPORT
 void loadpstoeditplugins(const char *progname, ostream & errstream, bool verbose)
 {
@@ -294,7 +310,7 @@ void loadpstoeditplugins(const char *progname, ostream & errstream, bool verbose
 		pluginsloaded |= loadPlugInDrivers(szExePath, errstream,verbose);
 	  }
 	  // now try also $exepath/../lib/pstoedit
-      if (szExePath[0]) {
+          if (szExePath[0]) {
          // it is not an empty string
 #if COMPILEDFOR64BIT 
 	    strcat_s(szExePath,1000,"/../lib64/pstoedit");
@@ -306,6 +322,24 @@ void loadpstoeditplugins(const char *progname, ostream & errstream, bool verbose
 	    }
 	  }
 	}
+#if !defined(_MSC_VER) && !defined(__CYGWIN__)
+	if (!pluginsloaded) {
+	  // finally try the directory where libpstoedit.so.0 was found
+	  // libpstoedit.so.0
+          RSString pathname_of_libpstoedit("");
+          int found = dl_iterate_phdr( dl_iterate_callback, &pathname_of_libpstoedit);
+	  if (found && pathname_of_libpstoedit != "") {
+	    char * fullfilename = cppstrdup(pathname_of_libpstoedit.c_str());
+	    char * dirname = strrchr(fullfilename, directoryDelimiter);
+	    if (dirname) {
+		*dirname = '\0';
+		pluginsloaded |= loadPlugInDrivers(fullfilename, errstream,verbose);
+	    }
+	    delete [] fullfilename;
+	  }
+	}
+#endif
+
 #ifdef PSTOEDITLIBDIR
 	if (!pluginsloaded) {
 		struct stat s;
@@ -365,6 +399,14 @@ ProgramOptions* getProgramOptions() {
 	return &PsToEditOptions::theOptions();
 }
 
+static RSString suffixOfFilename(const char* filename) {
+	const char* suffix = strrchr(filename,'.');
+	if (suffix) {
+		return RSString(suffix + 1);
+	} else {
+		return "";
+	}
+}
 
 extern "C" DLLEXPORT
 int pstoedit(int argc, const char *const argv[], ostream & errstream,
@@ -441,7 +483,7 @@ int pstoedit(int argc, const char *const argv[], ostream & errstream,
 		errstream "built: " << __DATE__ << " - " 
 #endif
 		errstream << buildtype << " - " << compversion << ")"
-		" : Copyright (C) 1993 - 2023 Wolfgang Glunz\n";
+		" : Copyright (C) 1993 - 2024 Wolfgang Glunz\n";
 	}
 
 	//  handling of derived parameters
@@ -566,8 +608,6 @@ int pstoedit(int argc, const char *const argv[], ostream & errstream,
 			commandline.addarg("-dDELAYBIND");
 		}
 		commandline.addarg("-dWRITESYSTEMDICT");
-
-
 		if (generalOptions.verbose()) {
 			commandline.addarg("-dESTACKPRINT");
 			//commandline.addarg("-sDEBUG=true");
@@ -586,7 +626,6 @@ To set the page size to match the EPS bounding box, use -dEPSCrop. To scale an E
 To get the pre 8.00 behaviour, either use -dNOEPS or run the file with (filename) (r) file cvs exec not (filename) run.
 */
 
-
 		for (unsigned int psi = 0; psi < generalOptions.psArgs().argc; psi++) {
 			if (strlen(generalOptions.psArgs().argv[psi])) {
 				commandline.addarg(generalOptions.psArgs().argv[psi]);
@@ -609,25 +648,26 @@ To get the pre 8.00 behaviour, either use -dNOEPS or run the file with (filename
 		}
 		return gsresult; // for gstest
 	}
+
 	if (generalOptions.drivername.value.length() == 0) {
-		// try to find driver according to suffix of input file
+		// try to find driver according to suffix of output file
 		if (!generalOptions.nameOfOutputFile) {
 			errstream << "No output format specified (-f option) and format could not be deduced from suffix of output file since no output filename was given" << endl;
 			shortusage(diag);
 			return 1;
 		} else {
-			const char * suffixOfInputFile = strrchr(generalOptions.nameOfOutputFile,'.');
-			if (!suffixOfInputFile) {
+			const RSString suffixOfOutputFile = suffixOfFilename(generalOptions.nameOfOutputFile);
+			if (suffixOfOutputFile == "") {
 				errstream << "No output format specified (-f option) and format could not be deduced from suffix of output file since no suffix was found" << endl;
 				shortusage(diag);
 				return 1;
 			} else {
-				const DriverDescription *suffixDriverDesc = getglobalRp()->getDriverDescForSuffix((suffixOfInputFile+1)); // +1 == skip "."
+				const DriverDescription *suffixDriverDesc = getglobalRp()->getDriverDescForSuffix(suffixOfOutputFile.c_str()); 
 				if (suffixDriverDesc) {
 					errstream << "No explicit output format specified - using " << suffixDriverDesc->symbolicname << " as derived from suffix of output file" << endl;
 					generalOptions.drivername = suffixDriverDesc->symbolicname;
 				} else {
-					errstream << "No output format specified (-f option) and format could not be uniquely deduced from suffix " << suffixOfInputFile << " of output file" << endl;
+					errstream << "No output format specified (-f option) and format could not be uniquely deduced from suffix " << suffixOfOutputFile << " of output file" << endl;
 					// usage(errstream);
 					getglobalRp()->explainformats(diag); // ,true);
 					return 1;
@@ -641,13 +681,17 @@ To get the pre 8.00 behaviour, either use -dNOEPS or run the file with (filename
 	{
 		// extract driver specific options from part after the : in the driver name
 		RSString drivername (generalOptions.drivername.value); // make a copy because it gets modified below. 
-		char *driveroptions_string = strchr(const_cast<char*>(drivername.c_str()), ':'); //lint !e1773
-		if (driveroptions_string) {
-			*driveroptions_string = '\0';	// replace : with 0 to separate drivername
-			// copy back the drivername - above we "hacked" into the string without adjusting internal state
-			const RSString tmp(drivername.c_str());
-			drivername = tmp;
-			driveroptions_string++;
+		RSString driveroptions("");
+		{
+			char *drivername_copy = cppstrdup(drivername.c_str());
+			char *driveroptions_string = strchr(drivername_copy, ':'); //lint !e1773
+			if (driveroptions_string) {
+				*driveroptions_string = '\0';	// replace : with 0 to separate drivername
+				driveroptions_string++;
+				driveroptions = RSString(driveroptions_string); // part post :
+				drivername    = RSString(drivername_copy);  // first part
+			}
+			delete [] drivername_copy;
 		}
 		//errstream << "DEBUG: " << drivername << endl;
 
@@ -662,7 +706,7 @@ To get the pre 8.00 behaviour, either use -dNOEPS or run the file with (filename
 			errstream << "The driver for the selected format cannot write to standard output because it writes binary data" << endl;
 			return 1;
 		}
-		if (driveroptions_string && strequal(driveroptions_string, "-help") ) {
+		if (driveroptions == "-help") {
 			ProgramOptions* dummy = currentDriverDesc->createDriverOptions();
 			if (dummy->numberOfOptions() ) {
 				dummy->showhelp(diag, "The driver for this output format supports the following additional options: "
@@ -707,25 +751,25 @@ To get the pre 8.00 behaviour, either use -dNOEPS or run the file with (filename
 			RSString gsdevice;
 			if (is_gs_shortcut) {
 				// extract gs driver from part after the '-' in the driver name if available
-				RSString drivernamecopy(drivername);
-				char* gsdevice_string = strchr(const_cast<char*>(drivernamecopy.c_str()), '-'); //lint !e1773
+				const char* gsdevice_string = strchr(drivername.c_str(), '-'); //lint !e1773
 				if (gsdevice_string) {
 					gsdevice = RSString(gsdevice_string + 1);
-					if (drvbase::Verbose()) {
-						errstream << "using gs device:" << gsdevice << endl;
-					}
 				} else {
 					errstream << "unexpected flow in " << __FILE__ << ":" << __LINE__ << endl;
 					assert(false);
 					return 1;
 				}
 			} else {
-				if (!driveroptions_string) {
+				if (driveroptions == "") {
 					errstream <<
 						"The gs output driver needs a gs-device as argument, e.g. gs:pdfwrite" << endl;
 					return 1;
 				}
-				gsdevice = RSString(driveroptions_string);
+				//errstream << "DEBUG: driveroptions: " << driveroptions << endl;
+				gsdevice = RSString(driveroptions.c_str());
+			}
+			if (drvbase::Verbose()) {
+				errstream << "using gs device:" << gsdevice << endl;
 			}
 			assert(gsdevice != "");
 			// special handling of direct Ghostscript drivers
@@ -751,6 +795,7 @@ To get the pre 8.00 behaviour, either use -dNOEPS or run the file with (filename
 			commandline.addarg("-dBATCH");
 			RSString tempbuffer("-sDEVICE=") ; 
 			tempbuffer += gsdevice; // e.g., pdfwrite ;
+			// we shouldn't need the --permit-devices since we have the -sDEVICE= option
 			commandline.addarg(tempbuffer.c_str() );
 			for (unsigned int psi = 0; psi < generalOptions.psArgs().argc; psi++) {
 				if (strlen(generalOptions.psArgs().argv[psi])) {
@@ -774,6 +819,7 @@ To get the pre 8.00 behaviour, either use -dNOEPS or run the file with (filename
 					      << " executed command : " << commandline << endl;
 			}
 			return gsresult;
+			// end case for gs devices
 		} else {
 			//
 			// standard case - neither gstest, nor a gs device
@@ -860,7 +906,7 @@ To get the pre 8.00 behaviour, either use -dNOEPS or run the file with (filename
 					}			// fail
 				}				// backend opens file by itself
 				outputdriver =
-					currentDriverDesc->CreateBackend(driveroptions_string,
+					currentDriverDesc->CreateBackend(driveroptions.c_str(),
 													 *outputFilePtr,
 													 errstream,
 													 generalOptions.nameOfInputFile,
@@ -874,7 +920,7 @@ To get the pre 8.00 behaviour, either use -dNOEPS or run the file with (filename
 				} else {
 					outputFilePtr = &cout;
 					outputdriver =
-						currentDriverDesc->CreateBackend(driveroptions_string,
+						currentDriverDesc->CreateBackend(driveroptions.c_str(),
 														 *outputFilePtr,
 														 errstream,
 														 generalOptions.nameOfInputFile,
@@ -982,6 +1028,9 @@ To get the pre 8.00 behaviour, either use -dNOEPS or run the file with (filename
 				if (generalOptions.usePlainStrings) {
 					inFileStream << "/pstoedit.ashexstring false def" << endl;
 				}
+				if (generalOptions.useFindDevice) {
+					inFileStream << "/pstoedit.usefinddevice true def" << endl;
+				}
 				if (generalOptions.t2fontsast1) {
 					inFileStream << "/pstoedit.t2fontsast1 true def" << endl;
 				} else {
@@ -1103,6 +1152,54 @@ To get the pre 8.00 behaviour, either use -dNOEPS or run the file with (filename
 					writeFileName(inFileStream, generalOptions.nameOfIncludeFile.value.c_str());
 					inFileStream << ") def" << endl;
 				}
+				int gs_revision = 0; 
+				{
+					// retrieve revision of ghostscript
+					Argv commandline;
+					const char* gstocall = whichPI_func(errstream, generalOptions.verbose(),
+						generalOptions.gsregbase.value.c_str(), generalOptions.GSToUse.value.c_str());
+					if (gstocall == nullptr) {
+						return 3;
+					}
+					commandline.addarg(gstocall);
+					// -dNODISPLAY -sstdout=v.txt -q -c revision 100 string cvs print quit
+					commandline.addarg("-dNODISPLAY");
+					commandline.addarg("-sstdout=" + gsout);
+					commandline.addarg("-q");
+					commandline.addarg("-c");
+					commandline.addarg("revision");
+					commandline.addarg("100");
+					commandline.addarg("string");
+					commandline.addarg("cvs");
+					commandline.addarg("print");
+					commandline.addarg("quit");
+
+					if (generalOptions.verbose()) {
+						errstream << "now calling the interpreter via: " << commandline << endl;
+					}
+					gsresult = call_PI_func(commandline.argc, commandline.argv);
+					if (gsresult != 0)  {
+						errstream << "PostScript/PDF Interpreter finished. Return status " << gsresult
+							<< " executed command : " << commandline << endl;
+					}
+					ifstream versionfile(gsout);
+					
+					versionfile >> gs_revision;
+					if (generalOptions.verbose()) {
+						errstream << "Revision of PostScript Interpreter: " << gs_revision << endl;
+					}
+				}
+
+				const RSString suffixOfInputFile = suffixOfFilename(generalOptions.nameOfInputFile);
+				if ((gs_revision > 9561 ) &&
+					((suffixOfInputFile == "pdf") || (suffixOfInputFile == "PDF")) ) {
+			       // since version 9.56.1 of GhostScript, the PDF interpreter is completely written in C and not longer works with the PostScript
+				   // level hooking that pstoedit uses
+					errstream << "Warning: PDF files cannot be converted anymore when using versions of GhostScript newer than 9.56.1." << endl;
+					errstream << "The generated output file will contain no image data." << endl;
+				}
+
+				const char * permit_devices = nullptr;
 
 				switch (currentDriverDesc->backendDesiredImageFormat) {
 					case DriverDescription::imageformat::noimage :
@@ -1111,18 +1208,21 @@ To get the pre 8.00 behaviour, either use -dNOEPS or run the file with (filename
 							inFileStream << "/pstoedit.backendSupportsFileImages true def" << endl;
 							inFileStream << "/pstoedit.withimages true def" << endl;
 							inFileStream << "/pstoedit.imagedevicename (bmp16m) def" << endl;
+							permit_devices="--permit-devices=bmp16m";
 							inFileStream << "/pstoedit.imagefilesuffix (.bmp)   def" << endl;
 							break;
 					case DriverDescription::imageformat::png :
 							inFileStream << "/pstoedit.backendSupportsFileImages true def" << endl;
 							inFileStream << "/pstoedit.withimages true def" << endl;
 							inFileStream << "/pstoedit.imagedevicename (png16m) def" << endl;
+							permit_devices="--permit-devices=png16m";
 							inFileStream << "/pstoedit.imagefilesuffix (.png)   def" << endl;
 							break;
 					case DriverDescription::imageformat::eps :
 							inFileStream << "/pstoedit.backendSupportsFileImages true def" << endl;
 							inFileStream << "/pstoedit.withimages true def" << endl;
 							inFileStream << "/pstoedit.imagedevicename (epswrite) def" << endl;
+							permit_devices="--permit-devices=epswrite";
 							inFileStream << "/pstoedit.imagefilesuffix (.eps)   def" << endl;
 							break;
 					case DriverDescription::imageformat::memoryeps :
@@ -1188,7 +1288,7 @@ To get the pre 8.00 behaviour, either use -dNOEPS or run the file with (filename
 					inFileStream << "/pstoedit.preps2ai where not { \n"
 						// first run through this file (before ps2ai.ps)
 						"	/jout true def \n" "	/joutput pstoedit.outputfilename def \n";
-					if (driveroptions_string && strequal(driveroptions_string, "-88")) {
+					if (driveroptions == "-88") {
 						inFileStream << "	/jtxt3 false cdef\n";
 					}
 					inFileStream << "	/pstoedit.textastext where { pop pstoedit.textastext not {/joutln true def } if } if \n" "	/pstoedit.preps2ai false def \n" "}{ \n" "	pop\n"	// second run (after ps2ai.ps)
@@ -1242,6 +1342,10 @@ To get the pre 8.00 behaviour, either use -dNOEPS or run the file with (filename
 				if (pioptions && (strlen(pioptions) > 0)) {
 					commandline.addarg(pioptions);
 				}
+				if (permit_devices && (gs_revision >= 10040)) {
+					// since 10.04.0 we need to add the permit_devices option. Earlier this is no known.
+					commandline.addarg(permit_devices);
+				}
 				if (!generalOptions.verbose())
 					commandline.addarg("-q");
 				if (drivername != "ps2ai") {	// not for ps2ai
@@ -1273,6 +1377,9 @@ To get the pre 8.00 behaviour, either use -dNOEPS or run the file with (filename
 					if (generalOptions.pngimage.value.length()) {
 						commandline.addarg("-dNOPAUSE");
 						commandline.addarg("-dBATCH");
+						if (gs_revision >= 10400) {
+							commandline.addarg("--permit-devices=png16m");
+						}
 						commandline.addarg("-sDEVICE=png16m");
 						RSString tempbuffer = "-sOutputFile=";
 						tempbuffer += generalOptions.pngimage.value;
@@ -1284,10 +1391,6 @@ To get the pre 8.00 behaviour, either use -dNOEPS or run the file with (filename
 				commandline.addarg("-dDELAYSAFER");
 				commandline.addarg("-dNOEPS"); // otherwise EPSF files create implicit showpages and a save/restore pair which disturbs the setPageSize handling
 
-				//layer commandline.addarg("-dDEBUG");
-				//layer commandline.addarg("pdf_cslayer.ps");
-
-
 				for (unsigned int psi = 0; psi < generalOptions.psArgs().argc; psi++) {
 					if (strlen(generalOptions.psArgs().argv[psi])) {
 						commandline.addarg(generalOptions.psArgs().argv[psi]);
@@ -1308,9 +1411,9 @@ To get the pre 8.00 behaviour, either use -dNOEPS or run the file with (filename
 				} else {
 					commandline.addarg(gsinfilename.c_str());
 				}
-				if (generalOptions.verbose())
+				if (generalOptions.verbose()) {
 					errstream << "now calling the interpreter via: " << commandline << endl;
-				// gsresult = system(commandline);
+				}
 				gsresult = call_PI_func(commandline.argc, commandline.argv);
 				if (gsresult != 0) {
 					errstream << "PostScript/PDF Interpreter finished. Return status " << gsresult
@@ -1378,7 +1481,7 @@ To get the pre 8.00 behaviour, either use -dNOEPS or run the file with (filename
 									  driverOptions,
 									  generalOptions.nameOfInputFile,
 									  generalOptions.nameOfOutputFile,
-									  currentDriverDesc, driveroptions_string, generalOptions.splitpages, outputdriver);
+									  currentDriverDesc, driveroptions.c_str(), generalOptions.splitpages, outputdriver);
 					//	if (options.verbose())
 					//		errstream << "now reading BoundingBoxes from file " << bbfilename << endl;
 						/* outputdriver-> */ 
